##############################################################################
#
#    Accounting Reports, for Odoo
#    Copyright © 2015 XCG Consulting <https://xcg-consulting.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import time
import uuid

import odoo.exceptions
import odoo.tests

# Default amount set on accounting documents.
AMOUNT = 42.0


@odoo.tests.tagged("-at_install", "post_install")
class Test(odoo.tests.TransactionCase):
    """Test this ``account_report`` module."""

    def setUp(self):
        """Set up elements we are going to use here."""

        super().setUp()

        self.sales_journal = self.env["account.journal"].search(
            [("type", "=", "sale"), ("company_id", "=", self.env.company.id)], limit=1
        )
        self.assertTrue(self.sales_journal)

        self.pay_journal = self.env["account.journal"].search(
            [("type", "=", "bank"), ("company_id", "=", self.env.company.id)], limit=1
        )
        self.assertTrue(self.pay_journal)

        self.debit_account = self._createAccount("asset_receivable")
        self.credit_account = self._createAccount("income_other")

        self.product = self.env.ref("product.product_product_1")

        self.a_period = self._create_account_period()

    def test_consult_account(self):
        """Consult an account."""

        account = self.debit_account

        # Create an accounting document.
        accdoc = self._makeAccDoc()

        # Find the accounting entry we expect to be extracted.
        accentry = self.env["account.move.line"].search(
            [("account_id", "=", account.id), ("move_id", "=", accdoc.id)]
        )
        self.assertEqual(len(accentry), 1)

        # Default export.
        self.assertEqual(self._consult_account(account), accentry.ids)

        # Ask for drafts only; should include our draft entry.
        self.assertEqual(
            self._consult_account(account, transaction_state="draft"),
            accentry.ids,
        )

        # Validate the accounting document; we should still get it.
        self._validateAccDoc(accdoc)
        self.assertEqual(self._consult_account(account), accentry.ids)

        # Ask for drafts only; should now find nothing to export.
        with self.assertRaises(odoo.exceptions.UserError):
            self._consult_account(account, transaction_state="draft")

        # Close the period; we should still find our entry as closed periods
        # are included by default.
        accentry.period_id.state = "done"
        self.assertEqual(self._consult_account(account), accentry.ids)

        # Ask for open periods only; should now find nothing to export.
        with self.assertRaises(odoo.exceptions.UserError):
            self._consult_account(account, include_closed_periods=False)

    def test_invoice_paid_amount(self):
        """Check the "Paid amount" field we have added into invoices."""

        # Create an invoice.
        client = self._createPartner()
        invoice = self._makeInvoice(client)
        self.assertEqual(invoice.paid_amount, 0.0)  # No payments so far.

        # Validate the invoice.
        self.assertEqual(invoice.state, "draft")
        self._validateAccDoc(invoice)
        invoice.invalidate_model()
        self.assertEqual(invoice.state, "posted")
        self.assertEqual(invoice.paid_amount, 0.0)  # No payments so far.

        # Partial payment.
        self._payInvoice(invoice, 2.0)
        invoice.invalidate_model()
        self.assertEqual(invoice.paid_amount, 2.0)
        self.assertEqual(invoice.state, "posted")

        # Partial payment.
        self._payInvoice(invoice, 4.0)
        invoice.invalidate_model()
        self.assertEqual(invoice.paid_amount, 2.0 + 4.0)
        self.assertEqual(invoice.state, "posted")

        # Pay the rest.
        self._payInvoice(invoice, invoice.amount_total - 2.0 - 4.0)
        invoice.invalidate_model()
        self.assertAlmostEqual(invoice.paid_amount, invoice.amount_total, places=2)
        self.assertEqual(invoice.payment_state, "paid")

    def _consult_account(self, account, **kwargs):
        """Consult the specified account; return extracted accounting entries.

        :type account: Odoo "account.account" record set.
        :param **kwargs: Account consultation dialog box creation values.
        :rtype: Odoo "account.move.line" ID list.
        """

        aconsult_dlg_values = {"account_id": account.id}
        aconsult_dlg_values.update(kwargs)
        aconsult_dlg = self.env["account.consultation"].create(aconsult_dlg_values)
        self.assertTrue(aconsult_dlg.account_type_ids)
        dlg_ret = aconsult_dlg.open_general_list()

        # The dialog box returns an action to display accounting entries; pick
        # them from there.
        # 'domain': [('id', 'in', amls.ids)],
        return dlg_ret["domain"][0][2]

    def _createAccount(self, account_type):
        """Create an account.

        :type account_type: String.
        :rtype: Odoo "account.account" record.
        """

        return self.env["account.account"].create(
            {
                "code": uuid.uuid4().hex[:8],
                "name": uuid.uuid4().hex,
                "reconcile": True,
                "account_type": account_type,
            }
        )

    def _createPartner(self):
        """Create a client.

        :rtype: Odoo "res.partner" record.
        """

        return self.env["res.partner"].create(
            {
                "is_company": True,
                "name": "TEST-PARTNER-NAME",
                "property_account_receivable_id": self.debit_account.id,
            }
        )

    def _makeAccDoc(self):
        """Create an accounting document and return it.

        :rtype: Odoo "account.move" record set.
        """

        # Prepare accounting entry creation values.
        accentry_values = [
            {  # debit
                "account_id": self.debit_account.id,
                "credit": 0.0,
                "date_maturity": odoo.fields.Date.today(),
                "debit": AMOUNT,
                "name": uuid.uuid4().hex,
            },
            {  # credit
                "account_id": self.credit_account.id,
                "credit": AMOUNT,
                "date_maturity": odoo.fields.Date.today(),
                "debit": 0.0,
                "name": uuid.uuid4().hex,
            },
        ]

        # Create and return a draft accounting document.
        accdoc = self.env["account.move"].create(
            {
                "journal_id": self.sales_journal.id,
                "line_ids": [(0, 0, values) for values in accentry_values],
                "period_id": self.a_period.id,
                "ref": uuid.uuid4().hex[:16],
            }
        )

        # Check a field we have added.
        for accentry in accdoc.line_ids:
            if accentry.account_id.id == self.debit_account.id:
                self.assertEqual(accentry.reporting_balance, AMOUNT)
            elif accentry.account_id.id == self.credit_account.id:
                self.assertEqual(accentry.reporting_balance, -AMOUNT)
        self.assertEqual(sum(accdoc.line_ids.mapped("reporting_balance")), 0.0)

        return accdoc

    def _makeInvoice(self, client):
        """Create an accout.move with account.move.line and return it.

        :type client: Odoo "res.partner" record.

        :return: The account entry representing the invoice.
        :rtype: Odoo "account.move" record.
        """

        return self.env["account.move"].create(
            {
                "partner_id": client.id,
                "move_type": "out_invoice",
                "invoice_line_ids": [
                    (
                        0,  # 0: Create.
                        0,
                        {
                            "account_id": self.credit_account.id,
                            "name": "TEST-INVLINE",
                            "price_unit": AMOUNT,
                            "product_id": self.product.id,
                            "quantity": 1.0,
                        },
                    )
                ],
            }
        )

    def _payInvoice(self, invoice, amount_to_pay):
        """Register a payment to pay the specified invoice. Partial payments
        supported.

        :type invoice: Odoo "account.move" record.

        :param amount_to_pay: How much of the invoice is being paid.
        :type amount_to_pay: Float.
        """

        acc_entry_to_pay = invoice.line_ids.filtered(
            lambda aml: aml.account_id.id == self.debit_account.id
        )
        self.assertEqual(len(acc_entry_to_pay), 1)

        wizard_obj = self.env["account.payment.register"].with_context(
            active_model="account.move.line",
            active_ids=acc_entry_to_pay.ids,
        )

        wizard = wizard_obj.create({"amount": amount_to_pay})
        wizard.action_create_payments()

        return True

    def _validateAccDoc(self, accdoc):
        """Validate the specified accounting document.

        :type accdoc: Odoo "account.move" record set.
        """

        self.assertEqual(accdoc.state, "draft")
        accdoc.action_post()
        self.assertEqual(accdoc.state, "posted")

    def _pay_partiallyAccDoc(self, accdoc):
        """Validate the specified accounting document.

        :type accdoc: Odoo "account.move" record set.
        """

        self.assertEqual(accdoc.state, "draft")
        accdoc.action_post()
        self.assertEqual(accdoc.state, "posted")

    def _create_account_period(self):
        """Create an accout.period and return it

        :rtype: Odoo "account.period" record.
        """

        fiscalyear = self.env["account.fiscal.year"].search(
            [
                ("date_from", "=", time.strftime("%Y") + "-01-01"),
                ("date_to", "=", time.strftime("%Y") + "-12-31"),
                ("company_id", "=", self.env.ref("base.main_company").id),
            ],
            limit=1,
        )

        if not fiscalyear:

            # Create an account.fiscal.year
            a_fiscalyear = self.env["account.fiscal.year"].create(
                {
                    "name": "Fiscal Year X " + time.strftime("%Y"),
                    "company_id": self.env.ref("base.main_company").id,
                    "date_from": time.strftime("%Y") + "-01-01",
                    "date_to": time.strftime("%Y") + "-12-31",
                }
            )

            # Create an account.period
            a_period = self.env["account.period"].create(
                {
                    "name": "Fiscal Year X " + time.strftime("%Y"),
                    "fiscalyear_id": a_fiscalyear.id,
                    "date_start": time.strftime("%Y") + "-01-01",
                    "date_stop": time.strftime("%Y") + "-12-31",
                    "company_id": self.env.ref("base.main_company").id,
                }
            )

        else:
            a_period = fiscalyear.period_ids[0]

        return a_period
