##############################################################################
#
#    Accounting Reports, for Odoo
#    Copyright © 2015 XCG Consulting <https://xcg-consulting.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from odoo import _, api, exceptions, fields, models


class AccountConsultation(models.TransientModel):
    """Dialog box to open accounting entries ("account.move.line" objects) of 1
    accounting account.
    """

    _name = "account.consultation"
    _description = "Account consultation dialog box"

    account_id = fields.Many2one(
        comodel_name="account.account",
        string="Accounting account",
        ondelete="cascade",
        help="The accounting account to consult.",
    )

    account_type_ids = fields.Many2many(
        comodel_name="account.consultation.type",
        string="Account categories",
        default=lambda self: self.env["account.consultation.type"].search([]),
        ondelete="cascade",
        help="Type used to filter selectable accounting accounts.",
    )

    due_date_from = fields.Date(
        string="Due date: From",
        help="The earliest due date to include transactions from.",
    )

    due_date_to = fields.Date(
        string="Due date: To",
        help="The latest due date to include transactions to.",
    )

    include_closed_periods = fields.Boolean(
        string="Include closed periods", default=True
    )

    include_reconciled = fields.Boolean(
        string="Include reconciled transactions", default=True
    )

    matching_count = fields.Integer(
        string="Matching entries - Count",
        help="Transactions with & without reference that match the filter.",
        readonly=True,
    )

    norefs_matching_count = fields.Integer(
        string="Matching entries without ref - Count",
        help="Transactions without reference that match the filter.",
        readonly=True,
    )

    partner_id = fields.Many2one(
        comodel_name="res.partner",
        string="Partner",
        ondelete="cascade",
        help="The partner to consult.",
    )

    @api.depends("include_closed_periods")
    def _get_period_closed_state(self):
        """Trick used in period_from_id / period_to_id view-side domains."""
        for record in self:
            record.period_closed_state = (
                "done" if record.include_closed_periods else "draft"
            )

    period_closed_state = fields.Char(
        compute=_get_period_closed_state,
        string="Period closed state (Utility)",
    )

    period_from_id = fields.Many2one(
        comodel_name="account.period",
        string="Period: From",
        help=(
            "The earliest accounting period to include transactions from. "
            "Both open and closed accounting periods may be selected here "
            'when the "Include closed periods" box is checked.'
        ),
        ondelete="set null",
    )

    period_to_id = fields.Many2one(
        comodel_name="account.period",
        string="Period: To",
        help=(
            "The latest accounting period to include transactions to. Both "
            "open and closed accounting periods may be selected here when the "
            '"Include closed periods" box is checked.'
        ),
        ondelete="set null",
    )

    ref_from = fields.Char(
        string="Reference: From",
        help="The first reference to include transactions from.",
    )

    ref_to = fields.Char(
        string="Reference: To",
        help="The last reference to include transactions to.",
    )

    refs_matching = fields.Char(
        string="Matching references",
        help="Transaction references that match the filter.",
        readonly=True,
    )

    refs_matching_count = fields.Integer(
        string="Matching reference count",
        help="Transaction references that match the filter.",
        readonly=True,
    )

    transaction_state = fields.Selection(
        selection=[
            ("draft", "Draft"),
            ("validated", "Validated"),
            ("all", "All"),
        ],
        string="State",
        default="all",
        help="Filter based on transaction states.",
        required=True,
    )

    @api.onchange("account_type_ids", "partner_id")
    def _handle_account_criteria_change(self):
        """Update selectable accounts based on criteria."""

        partner = self.partner_id

        if self.account_id.account_type not in self.account_type_ids.mapped(
            "account_type"
        ):
            self.account_id = self.env["account.account"]  # Clear out.

        account_domain = (
            [("account_type", "in", self.account_type_ids.mapped("account_type"))]
            if self.account_type_ids
            else []
        )

        if partner:
            accounts = self.env["account.account"]
            for partner_field in self.get_partner_account_fields():
                account = getattr(partner, partner_field)
                if account:
                    accounts |= account
            account_domain.append(("id", "in", accounts.ids))

        return {"domain": {"account_id": account_domain}}

    def remove_all_categories(self):
        self.account_type_ids = False

    @api.onchange(
        "account_id",
        "account_type_ids",
        "partner_id",
        "period_from_id",
        "period_to_id",
        "due_date_from",
        "due_date_to",
        "ref_from",
        "ref_to",
        "transaction_state",
        "include_closed_periods",
        "include_reconciled",
    )
    def _handle_criteria_change(self):
        """Update the ref preview when changing criteria."""

        self._refresh_ref_preview()

    def name_get(self):
        """Override to name this record as we display it inline."""

        ret = []

        for record in self:
            account = record.account_id

            # 0 & 1 below: name_get returns a list of (ID, name) tuples.
            ret.append(
                (
                    record.id,
                    _("Account consultation - %s")
                    % (
                        account.name_get()[0][1]
                        if account
                        else record.partner_id.name_get()[0][1]
                    ),
                )
            )

        return ret

    @api.model
    def find_account_types(self, account_types):

        return self.env["account.consultation.type"].search(
            [
                ("account_type", "in", account_types),
            ]
        )

    def get_partner_account_fields(self):
        """Provide partner fields that reference accounts. This method is here
        to allow overrides.
        :rtype: String list.
        """

        return [
            "property_account_receivable_id",
            "property_account_payable_id",
        ]

    def open_aml_list(self, list_view_ref, list_view_title):
        """Open the accounting entry list for the specified account.
        :param list_view_ref: The "account.move.line" view to display.
        :type list_view_ref: String.
        :param list_view_title: Title of the view to display.
        :type list_view_title: String.
        :return: Action to display accounting entries.
        :rtype: Dictionary.
        """

        self.ensure_one()

        amls = self.env["account.move.line"].search(self._build_aml_domain())

        if not amls:
            raise exceptions.UserError(
                _(
                    "No accounting entry could be found with the specified "
                    "parameters."
                )
            )

        self._refresh_ref_preview()

        self = self.with_context(account_consultation=True)

        # Show accounting entries.
        return {
            "context": self.env.context,
            "domain": [("id", "in", amls.ids)],
            "name": list_view_title,
            "res_model": "account.move.line",
            "type": "ir.actions.act_window",
            "view_mode": "tree,form",
            "views": [
                (self.env.ref(list_view_ref).id, "tree"),
                (self.env.ref("account.view_move_line_form").id, "form"),
            ],
        }

    def open_general_list(self):
        """Open the accounting entry list for the specified account."""

        self.ensure_one()

        return self.open_aml_list(
            "account_report.account_consultation_general_aml_list",
            _("General list"),
        )

    def open_analytic_list(self):
        """Open the accounting entry list for the specified account."""

        self.ensure_one()

        return self.open_aml_list(
            "account_report.account_consultation_analytic_aml_list",
            _("Analytic list"),
        )

    def print_statement(self):
        """Print an account statement report (which needs to have been set up
        in advance).
        :return: Action to print the report.
        :rtype: Dictionary.
        """

        self.ensure_one()

        amls = self.env["account.move.line"].search(self._build_aml_domain())

        if not amls:
            raise exceptions.UserError(
                _(
                    "No accounting entry could be found with the specified "
                    "parameters."
                )
            )

        self._refresh_ref_preview()

        # Build a context that points to our accounting entries.
        report_context = self.env.context.copy()
        report_context.update(
            {"active_ids": amls.ids, "active_model": "account.move.line"}
        )
        report_context.pop("active_id", None)

        # Let the reporting engine do the rest.
        return (
            self.env["ir.actions.report"]
            .search(
                [
                    (
                        "report_name",
                        "=",
                        "account_report.report_account_statement",
                    )
                ],
                limit=1,
            )
            .with_context(report_context)
            .report_action(amls, data=None)
        )

    def _build_aml_domain(self):
        """Return a domain built based on the specified parameters made to find
        accounting entries.
        """

        account = self.account_id
        partner = self.partner_id
        categories = self.account_type_ids
        if not account and not partner:
            return [("id", "=", 0)]  # Not enough criteria.

        aml_domain = []

        if account:
            aml_domain = [("account_id", "=", account.id)]
        elif partner:
            aml_domain = [("partner_id", "=", partner.id)]

        aml_domain.append(
            ("account_id.account_type", "in", categories.mapped("account_type"))
        )

        period_states = ["draft"]
        if self.include_closed_periods:
            period_states.append("done")
        period_domain = [("state", "in", period_states)]
        if self.period_from_id:
            period_domain.append(("date_start", ">=", self.period_from_id.date_start))
        if self.period_to_id:
            period_domain.append(("date_start", "<=", self.period_to_id.date_start))
        periods = self.env["account.period"].search(period_domain)
        aml_domain.append(("period_id", "in", periods.ids))

        if self.transaction_state == "draft":
            aml_domain.append(("parent_state", "=", "draft"))
        elif self.transaction_state == "validated":
            aml_domain.append(("parent_state", "=", "posted"))

        if not self.include_reconciled:
            aml_domain.append(("full_reconcile_id", "=", False))

        if self.due_date_from:
            aml_domain.append(("date_maturity", ">=", self.due_date_from))
        if self.due_date_to:
            aml_domain.append(("date_maturity", "<=", self.due_date_to))

        # By default we include entries with empty refs.
        if self.ref_from:
            aml_domain += [("ref", "!=", False), ("ref", ">=", self.ref_from)]
        if self.ref_to:
            aml_domain += [("ref", "!=", False), ("ref", "<=", self.ref_to)]

        return aml_domain

    def _refresh_ref_preview(self):
        """Update the ref preview."""

        aml_refs = [
            aml["ref"]
            for aml in self.env["account.move.line"].search_read(
                self._build_aml_domain(), ["ref"], order="ref ASC"
            )
        ]

        # Discard entries without ref here but update our counters.
        entry_count = ref_count = norefs_count = 0
        filtered_aml_refs = []
        for ref in aml_refs:
            entry_count += 1
            if ref:
                filtered_aml_refs.append(ref)
                ref_count += 1
            else:
                norefs_count += 1
        aml_refs = filtered_aml_refs

        self.matching_count = entry_count
        self.refs_matching_count = ref_count
        self.norefs_matching_count = norefs_count

        if not ref_count:
            self.refs_matching = _("Accounting entries found have no reference.")

        elif ref_count <= 6:
            self.refs_matching = ", ".join(aml_refs)

        else:
            # We only care abbout the first 3 and the last 3.
            first_3 = aml_refs[:3]
            last_3 = aml_refs[-3:]
            self.refs_matching = ", ".join(first_3) + ", [....], " + ", ".join(last_3)
